<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.3/2027974
 */

namespace Application\Permissions;

use Application\Config\ConfigManager;
use Application\Factory\InvokableService;
use Application\Helper\ArrayHelper;
use Application\Permissions\Exception\ForbiddenException;
use Interop\Container\ContainerInterface;
use Workflow\Model\IWorkflow;

/**
 * A class following the 'permissions' pattern to check configuration values to see if
 * access is allowed to certain features. Using the 'permissions' service directly implies
 * 'authenticated' so this class can enforce without requiring 'authenticated'.
 * @package Application\Permissions
 */
class ConfigCheck implements InvokableService
{
    protected $services = null;

    const WORKFLOW_ARE_NOT_ENABLED = 'Workflows are not enabled';

    /**
     * @inheritdoc
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        $this->services = $services;
    }

    /**
     * Ensures that all of the provided checks pass
     * @param array     $checks         Checks to perform. Valid values are 'workflow' and any allowed by the
     *                                  'permissions' module. Checks unknown to this class are passed on to
     *                                  'permissions'.
     * @throws ForbiddenException
     * @throws \Application\Config\ConfigException
     */
    public function enforce($checks)
    {
        $checks = (array) $checks;
        foreach ($checks as $check) {
            switch ($check) {
                case IWorkflow::WORKFLOW:
                    if (!ConfigManager::getValue(
                        $this->services->get(ConfigManager::CONFIG),
                        ConfigManager::WORKFLOW_ENABLED,
                        false
                    )) {
                        throw new ForbiddenException(self::WORKFLOW_ARE_NOT_ENABLED);
                    }
                    break;
                default:
                    $this->services->get(Permissions::PERMISSIONS)->enforce($check);
                    break;
            }
        }
    }

    /**
     * Determines whether a string matches any of the strings in a blacklist.
     * A direct regex version of each of the blacklist strings will be used for comparison.
     * NOTE: The only exception to this is when the blacklist string only contains characters
     *       in the set of alphanumeric characters plus '_' and '@'. In this case, the
     *       regex will be set to an exact match. If you want an exact match on any special
     *       characters, you will need to escape them.
     *
     * Examples:
     *       Exact :      $blacklist = ['jim']          - matches 'jim' but does not match 'bigjim' or 'bigjimbob'
     *       Contains:    $blacklist = ['.*jim.*']      - matches 'jim' as well as 'bigjim' and 'bigjimbob'
     *       Contains:    $blacklist = ['.?jim.?']      - matches 'jim' as well as 'ajim' and 'ajimb'
     *       Starts with: $blacklist = ['jim.*']        - matches 'jim' and 'jimbob' but not 'bigjim' or 'bigjimbob'
     *       Ends with:   $blacklist = ['.*bob']        - matches 'jimbob' and 'bigjimbob' but not 'jim'
     *       With escape: $blacklist = ['jim\-bob\.x]   - matches 'jim-bob.x', only
     *
     * @param  string      $id                 The string that is to be checked for blacklisting
     * @param  array       $blacklist          A list of terms to compare the $id against
     * @param  bool        $caseSensitive      Whether the regex should be case sensitive
     *
     * @return bool
     */
    public static function isBlacklisted($id, $blacklist, $caseSensitive = false)
    {
        if (!empty($blacklist) && ArrayHelper::findMatchingStrings($blacklist, [$id], $caseSensitive)) {
            return true;
        }

        return false;
    }
}
